/* Agent Tcl
   Bob Gray
   26 June 1995

   genConnection.cc

   This file implements the classes that handle message-oriented connections.

   Copyright (c) 1995, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "platCharact.h"
#include "platExclusion.h"
#include "platSigio.h"		// SigioTrap
#include "agentMachId.h"	// MachineId
#include "genAgentIO.h"		// AgentIO
#include "genDefaults.h"	// AGENT_UNIX_ENV and AGENT_TCPIP_ENV
#include "genLocation.h"	// LOCATION

#include "genAgent.h"
#include "genAgentConn.h"
#include "genDBuffer.h"
#include "genEncrypt.h"
#include "genMasks.h"
#include "genMessage.h"
#include "genTcpip.h"
#include "genTransmit.h"

    /* static data */

UINT_32 AgentConnection::m_port
	= 0;

char *AgentConnection::m_unixsocket
	= NULL;

AgentConnectionProxy::~AgentConnectionProxy ()
{
    if (m_connectionOwner) {
	delete (m_connection);
    }
}

AgentConnectionProxy *AgentConnectionProxy::makeConnection 
    (struct timeval stop, 
     AGENT *agent, 
     const MachineId& machine, 
     const BASE_SECURITY &security,
     int &errorCode,
     AgentConnectionProxy::BackgroundHandler backgroundHandler
    )
{
    int rc;
    AgentIO *io;
    LOCATION *location;
    AgentConnection *connection;
    AgentConnectionProxy *proxy;
    MachineId scratchMachine = machine;
	
	/* background connection is not always allowed */

    if (backgroundHandler == e_BACKGROUND_OKAY) {

	    /* get the IP address of the machine */

        if ((rc = tcpip_getip (scratchMachine, stop)) == TCPIP_TIMEOUT) {
	    errorCode = MESSAGE_TIMEOUT;
	    return NULL;
        } else if (rc != TCPIP_OK) {
	    errorCode = MESSAGE_NOT_FOUND;
	    return NULL;
        }

	    /* get the AGENT_LOCATION */

        location = agent -> get_location ();
        assert (location != NULL);

	    /* get the actual IP address of the machine */

        if (scratchMachine.getIp() == location -> getLocalhostIp()) {
	    if (location -> getMachineIp() != UNKNOWN_IP) {
	        scratchMachine.setIp (location -> getMachineIp());
	    }
        }

	    /* see if we can use the background connection */

        if (scratchMachine.getIp() == location -> getLocalIp()) {

	    io = agent -> get_io();
	    assert (io != NULL);

	    if ((connection = io -> getBackgroundConnection ()) != NULL) {
	        proxy = new AgentConnectionProxy (connection);
	        return (proxy);
	    }
        }
    }

	/* otherwise we have to create a new connection */

    connection = new AgentConnection (agent, scratchMachine, security);

    if ((rc = connection -> connect (stop)) != MESSAGE_OK) {
	delete (connection);
	errorCode = rc;
	return (NULL);
    }

    proxy = new AgentConnectionProxy (connection, e_TRUE);
    errorCode = MESSAGE_OK;
    return (proxy);
}

AgentConnection::AgentConnection (AGENT *agent, const MachineId& machine, const BASE_SECURITY &security):
    m_agent (agent),
    m_machine (machine),
    m_security (security),
    m_connection (NULL),
    m_backgroundConnection (e_FALSE),
    m_trap (NULL),
    m_lock (NULL),
    m_locks (NULL),
    m_threads (PlatformCharacteristics::isMultiThreaded())
{
    assert (m_agent != NULL); 

#ifdef FIX_LATER
	/* we should have some kind of initialization routine to avoid */
	/* the race condition                                          */
#endif

	/* initialize the TCP/IP port number and the Unix socket filename */
	/* if necessary                                                   */

    if (m_unixsocket == NULL) {

	char *ptr;

	if ((ptr = getenv(AGENT_UNIX_ENV)) == NULL) {
	    m_unixsocket = UNIX_SOCKET;
	} else {
	    m_unixsocket = my_strcpy (ptr);
	}

	if ((ptr = getenv(AGENT_TCPIP_ENV)) == NULL) {
	    m_port = SERVER_TCP_PORT;
	} else if (sscanf (ptr, "%u", &m_port) < 1) {
	    fprintf (stderr, "\nFATAL ERROR: environment variable \"%s\" must contain an integer port number!\n\n", AGENT_TCPIP_ENV);
	    exit (1);
	}
    }

	/* initialize locking structures */

    if (m_threads) {
	m_lock = new MonitorLock ();
	m_locks = new MonitorGroup ();
	m_locks -> push (agent -> getMonitorLock());
	m_locks -> push (*m_lock);
    }
}

AgentConnection::AgentConnection (int fd, AGENT *agent, const MachineId& machine, const BASE_SECURITY &security):
    m_agent (agent),
    m_machine (machine),
    m_security (security),
    m_connection (NULL),
    m_backgroundConnection (e_FALSE),
    m_trap (NULL),
    m_lock (NULL),
    m_locks (NULL),
    m_threads (PlatformCharacteristics::isMultiThreaded())
{
    assert (m_agent != NULL);

    m_connection = new SocketConnection (fd, m_machine, m_security);

    if (m_threads) {
	m_lock = new MonitorLock ();
	m_locks = new MonitorGroup ();
	m_locks -> push (agent -> getMonitorLock());
	m_locks -> push (*m_lock);
    }
}
 
AgentConnection::~AgentConnection ()
{
    delete (m_connection);
}

BOOLEAN AgentConnection::isBackground (void)
{
    return (m_backgroundConnection);
}

void AgentConnection::makeBackground (ConnectionTrap *trap, MonitorGroup group)
{
	/* assertions */

    assert (m_connection != NULL);
    assert (m_trap == NULL);

	/* expand the set of monitor locks if necessary */

    if (m_threads) {
	m_lock -> acquire ();
	group.push (*m_lock);
    }

	/* turn on the background handler */

    m_trap = trap;
    m_connection -> turnBackgroundOn (this, group);
    m_backgroundConnection = e_TRUE;

	/* release our locks */

    if (m_threads) {
	m_lock -> release ();
    }
}

int AgentConnection::isMessagePending (int& flags)
{
    int rc;

    if (m_threads && m_backgroundConnection) {
	m_lock -> acquire ();
    }

    assert (m_connection != NULL);
    rc = m_connection -> checkReadyStatus (SigioTrap::e_READ_READY, flags);

    if (m_threads && m_backgroundConnection) {
	m_lock -> release();
    }

    return (rc);
}

int AgentConnection::connect (struct timeval stop)
{
    int rc;
    LOCATION *location;

	/* get the IP address of the machine */

    if ((rc = tcpip_getip (m_machine, stop)) == TCPIP_TIMEOUT) {
	return MESSAGE_TIMEOUT;
    } else if (rc != TCPIP_OK) {
	return MESSAGE_NOT_FOUND;
    }

	/* get the AGENT_LOCATION */

    location = m_agent -> get_location ();
    assert (location != NULL);

	/* get the actual IP address of the machine */

    if (m_machine.getIp() == location -> getLocalhostIp()) {
	if (location -> getMachineIp() != UNKNOWN_IP) {
	    m_machine.setIp (location -> getMachineIp());
	}
    }

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> acquireAll();
    }

	/* see if we can use the Unix domain socket */

    if (m_connection != NULL) {

	rc = MESSAGE_OK;

    } else if ((m_machine.getIp() == location -> getLocalhostIp()) || 
	       (m_machine.getIp() == location -> getMachineIp())) {

	UnixClientConnection *connection;
	connection = new UnixClientConnection (m_security);
	rc = connection -> connect (m_machine, m_unixsocket, stop);
	m_connection = connection;

    } else {

	TcpipClientConnection *connection;
        connection = new TcpipClientConnection (m_security);
	rc = connection -> connect (m_machine, m_port, stop);
	m_connection = connection;
    }

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_lock -> release();
    }

    return (rc);
}

int AgentConnection::send (MESSAGE &message)
{
    int rc;

	/* security is required only if we are sending over a new TCP/IP */
	/* link; security is *not* required if we are sending over the   */
	/* dedicated background connection				 */

    if (m_backgroundConnection != e_TRUE) {
	rc = m_connection -> send (message);
    } else {
	rc = m_connection -> unsecureSend (message);
    }

    return (rc);
}

int AgentConnection::unsecureSend (MESSAGE &message)
{
    int rc;

	/* security is required only if we are sending over a new TCP/IP */
	/* link; security is *not* required if we are sending over the   */
	/* dedicated background connection; in this case security just   */ 
	/* means sending an explicit indication that we do not consider  */
	/* security necessary                                            */

    if (m_backgroundConnection != e_TRUE) {

	MESSAGE securityMessage (protectionMessages.messages[PROT_NONE]);
	
	if ((rc = m_connection -> unsecureSend (securityMessage)) != MESSAGE_OK) {
	    return (rc);
	}
    }

    rc = m_connection -> unsecureSend (message);
    return (rc);
}

MESSAGE *AgentConnection::unsecureReceiveNoncomm (const MESSAGE_SET& set)
{
	/* get the masks */

    AgentMasks *masks = m_agent -> get_masks ();
    assert (masks != NULL);

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> acquireAll();
    }

	/* loop until we have the desired message -- handle any RESP_EVENT, */
	/* RESP_MEETING or RESP_MESSAGE messages that are received *before* */
	/* the desired message arrives                                      */

    MESSAGE *response;

    while (1) {

	response = m_connection -> unsecureReceive (set);

	if (response == NULL) {
	    break;
	}

	UINT_8 flag = response -> getFlag ();

	if ((flag != RESP_MEETING) && (flag != RESP_MESSAGE) && (flag != RESP_EVENT)) {
	    break;
	}

	masks -> handleIncoming_noLock (response);
	delete (response);
	response = NULL;
    }

	/* handle any remaining RESP_EVENT, RESP_MESSAGE or RESP_MEETING     */
	/* messages -- this is somewhat critical since we will not get a     */
	/* SIGIO alarm for any messages that arrived while we were receiving */
	/* previous messages                                                 */

    int flags;
    MESSAGE *otherResponse;

    while (!isMessagePending(flags) && (flags & SigioTrap::e_READ_READY)) {

	    /* receive the message */

	otherResponse = m_connection -> unsecureReceive (set);

	    /* break if we have no message */

	if (otherResponse == NULL) { 
	    break;
	}

	    /* fatal system bug if we did not get a RESP_MESSAGE, RESP_EVENT */
	    /* or RESP_MEETING message                                       */

	UINT_8 flag = otherResponse -> getFlag ();

	if ((flag != RESP_MEETING) && (flag != RESP_MESSAGE) && (flag != RESP_EVENT)) {
	    abort_with_message ("expected (but did not get) a RESP_MEETING, RESP_EVENT or RESP_MESSAGE message");
	}

	    /* handle the message */
 
	masks -> handleIncoming_noLock (otherResponse);
	delete (otherResponse);
	otherResponse = NULL;
    }

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> releaseAll();
    }

    return (response);
}

MESSAGE *AgentConnection::unsecureReceiveAny (const MESSAGE_SET& set)
{
    MESSAGE *response;

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_lock -> acquire ();
    }

	/* receive an incoming message */

    response = m_connection -> unsecureReceive (set);

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_lock -> release();
    }

	/* done */
	    
    return (response);
}

int AgentConnection::RPC (MESSAGE &message, const MESSAGE_SET& messageSet, MESSAGE*& response)
{
    int code;
    int flags;

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> acquireAll();
    }

	/* send message and receive response */

    if ((code = send (message)) == MESSAGE_OK) {

	    /* if sent successfully, receive the response */

	response = unsecureReceiveNoncomm (messageSet);

    } else {

	    /* if not sent successfully, check if their is a response and   */
	    /* then receive the response if present, since the server might */
	    /* have sent a response and closed the connection without even  */
	    /* receiving our request                                        */

	if (!isMessagePending(flags) && (flags & SigioTrap::e_READ_READY)) {

	    if ((response = unsecureReceiveNoncomm (messageSet)) != NULL) {
		code = MESSAGE_OK;
	    }

	} else {

	    response = NULL;
	}
    } 

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> releaseAll();
    }

    return (code);
}

int AgentConnection::unsecureRPC (MESSAGE &message, const MESSAGE_SET &messageSet, MESSAGE*& response)
{
    int code;
    int flags;

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> acquireAll();
    }

	/* send message and receive response */

    if ((code = unsecureSend (message)) == MESSAGE_OK) {

	    /* if sent successfully, receive the response */

	response = unsecureReceiveNoncomm (messageSet);

    } else {

	    /* if not sent successfully, check if their is a response and   */
	    /* the nreceive the response if present, since the server might */
	    /* have sent a response and closed the connection without       */
	    /* even receiving our request                                   */

	if (!isMessagePending(flags) && (flags & SigioTrap::e_READ_READY)) {

	    if ((response = unsecureReceiveNoncomm (messageSet)) != NULL) {
		code = MESSAGE_OK;
	    }

	} else {

	    response = NULL;
	}
    } 

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_locks -> releaseAll();
    }

    return (code);
}

int AgentConnection::unsecureSendFromBuffer (DynamicBuffer& buffer)
{
    int code;

	/* lock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_lock -> acquire();
    }

	/* do the send */

    code = m_connection -> unsecureSendFromBuffer (buffer);

	/* unlock if necessary */

    if (m_threads && m_backgroundConnection) {
	m_lock -> release();
    }

    return (code);
}

void AgentConnection::connectionAlarm (Connection &connection, Connection::ConnStatus status)
{
	/* NOTE: we don't need to acquire the lock since it was */
	/* acquired by the lower-level routines before making   */
	/* the callback				                */

	/* do nothing unless this is a read */

    if (status != Connection::e_READ_READY) {
	return;
    }

	/* make our callback */

    assert (m_trap != NULL);
    m_trap -> connectionAlarm (connection, status);
}
