/* Agent Tcl
   Bob Gray
   26 June 1995

   genConnection.h

   This file defines classes that handle message-oriented connections.

   Copyright (c) 1995, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef _GEN_CONNECTION_H
#define _GEN_CONNECTION_H

#ifndef NO_PRAGMAS
#pragma interface
#endif

#ifndef _PLAT_PORTING_H
#include "platPorting.h"	// fd_mask and abort_with_message
#endif
#ifndef _PLAT_SIGIO_H
#include "platSigio.h"		// class SigioTrap and SigioHandler
#endif
#ifndef _PLAT_SIZES_H
#include "platSizes.h"		// UINT_32
#endif
#ifndef _AGENT_ID_H
#include "agentId.h"
#endif
#ifndef _GEN_DSTRING_H
#include "genDString.h"		// class DynamicString
#endif
#ifndef _TRUEFALSE_H
#include "truefalse.h"		// BOOLEAN
#endif

    /* forward declaration to allow compilation */

class BASE_SECURITY;
class Connection;
class DynamicBuffer;
class MESSAGE;
class MESSAGE_SET;
class MonitorGroup;

    // file access callbacks

class FileCallbacks
{ 
    public:

	virtual int read 
	    (int n, char *buffer) = 0;

	virtual int write
	    (int n, const char *buffer) = 0;
};

    /* general connection */

class Connection
{
    public:

	enum ConnStatus {
	    e_INITIALIZED,
	    e_WAITING_TO_CONNECT,
	    e_WAITING_TO_ACCEPT,
	    e_ACCEPT_READY,
	    e_CONNECTED,
	    e_READ_READY,
	    e_WRITE_READY,
	    e_BOTH_READY,
	    e_FAILED
	};

	    // constructor and destructor

	Connection (const BASE_SECURITY &p_security);
	virtual ~Connection ();

	    /* see if the connection is available for reading and writing */

	virtual int checkReadyStatus 
	    (int inFlags, int &outFlags) = 0;

	    /* send a message along the connection */ 

	virtual int send
	    (MESSAGE &message) = 0;
	
	virtual int unsecureSend
	    (MESSAGE &message) = 0;

	virtual int unsecureSendFromBuffer
	    (DynamicBuffer &buffer) = 0;

	    /* receive a message from the connection */

	virtual MESSAGE * unsecureReceive
	    (const MESSAGE_SET &messageSet) = 0;

	virtual int unsecureReceiveIntoBuffer
	    (DynamicBuffer& buffer) = 0;

	    /* read and write from the connection */

	virtual int read 
	    (int n, char *buffer) = 0;

	virtual int write
	    (int n, const char *buffer) = 0;

	    /* get the file descriptor associated with connection if there is one */

	virtual int getFd (void) = 0;

    protected:

	    // get and set the status

	void setStatus (ConnStatus status) {
	    m_status = status;
	}

	ConnStatus getStatus (void) {
	    return (m_status);
	}

	BASE_SECURITY m_security;    // security information


    private:

	ConnStatus m_status;
};

    // connection callback

class ConnectionTrap
{
    public:

	virtual void connectionAlarm 
	    (Connection &connection, Connection::ConnStatus status) = 0;
};

    /* file connection */

class FileConnection: public Connection
{
    protected:

	FileCallbacks &m_callbacks;      // callbacks for file access

    public:

	    /* constructor and virtual destructor */

	FileConnection (FileCallbacks &callbacks, const BASE_SECURITY &security);
	virtual ~FileConnection ();

	    /* see if the connection is available for reading and writing */

	int checkReadyStatus
	    (int inFlags, int &outFlags);

	    /* send a message along the connection */

	virtual int send
	    (MESSAGE &message);
	
	virtual int unsecureSend
	    (MESSAGE &message);

	virtual int unsecureSendFromBuffer
	    (DynamicBuffer &buffer);

	    /* receive a message from the connection */

	virtual MESSAGE * unsecureReceive
	    (const MESSAGE_SET &messageSet);

	virtual int unsecureReceiveIntoBuffer
	    (DynamicBuffer& buffer);

	    /* read and write from the connection */

	virtual int read
	    (int n, char *buffer);

	virtual int write
	    (int n, const char *buffer);

	    /* get the file descriptor associated with the connection */

	int getFd (void) {
#ifdef FIX_LATER
	    // we don't yet need this fd 
#else
	    abort_with_message ("FileConnection::getFd has not been implemented!");
	    return (-1);
#endif
	}
};

    /* asynchronous I/O connection */

class AsyncConnection: public Connection
{
    public:

	    /* turn on and off the background handler */

	virtual void turnBackgroundOn 
	    (ConnectionTrap *newTrap, MonitorGroup group) = 0;

	virtual void turnBackgroundOff 
	    (void) = 0;

	    /* constructor and virtual destructor */

	AsyncConnection (const BASE_SECURITY &security);
	virtual ~AsyncConnection (void);
};

    /* socket connection */

class SocketConnection: public AsyncConnection, public SigioTrap
{
    protected:

	ConnectionTrap *m_trap;	        // trap to call on asynchronous I/O
	MachineId m_machine;            // machine on the other end
	SigioHandler *m_handler;        // SIGIO handler
	int m_sockfd;		        // socket descriptor
	fd_mask m_set [MASK_SIZE];      // set for select
	fd_mask m_index;                // index of m_sockfd in set

	    // callback from lower-level routines

	void sigioAlarm
	    (int fd, int flags);

    public:

	    // constructor and virtual destructor 

	SocketConnection 
	    (const BASE_SECURITY &security);

 	SocketConnection
	    (int sockfd, const MachineId &machine, const BASE_SECURITY &security);
	
	virtual ~SocketConnection ();

	    // turn on and off the background handler 

	void turnBackgroundOn 
	    (ConnectionTrap *newTrap, MonitorGroup group);

	void turnBackgroundOff 
	    (void);

	    /* see if the connection is available for reading and writing */

	int checkReadyStatus 
	    (int inFlags, int &outFlags);

	    /* send a message along the connection */

	int send
	    (MESSAGE &message);
	
	int unsecureSend
	    (MESSAGE &message);

	virtual int unsecureSendFromBuffer
	    (DynamicBuffer &buffer);

	    /* receive a message from the connection */

	MESSAGE * unsecureReceive
	    (const MESSAGE_SET &messageSet);

	virtual int unsecureReceiveIntoBuffer
	    (DynamicBuffer& buffer);

	    /* read and write from the connection */

	int read 
	    (int n, char *buffer);

	int write
	    (int n, const char *buffer);

	    /* return the file descriptor associated with the connection */

        int getFd (void) {
	    return (m_sockfd);
	}
};

    /* client-side TCP/IP connection */

class TcpipClientConnection: public SocketConnection
{
    protected:

	UINT_32 m_port;		    // port on destination machine

    public:

	    // constructors and destructor

	TcpipClientConnection 
	    (const BASE_SECURITY &security);
	
	virtual ~TcpipClientConnection
	    ();

	    /* connect to the destination machine */

	int connect 
	    (const MachineId& machine, UINT_32 port, struct timeval stop);

	int asyncConnect
	    (const MachineId& machine, UINT_32 port);
};

    /* client-side Unix domain connection */

class UnixClientConnection: public SocketConnection
{
    protected:

	DynamicString m_socketFilename;		// Unix socket filename

    public:

	    /* constructor and destructor */

	UnixClientConnection 
	   (const BASE_SECURITY &security);

       ~UnixClientConnection 
	   ();

	    /* connect to the destination machine */

	int connect 
	    (const MachineId& machine, const DynamicString& socketFilename, struct timeval stop);
};

    /* server-side TCP/IP connection */

class TcpipServerConnection: public SocketConnection
{
    protected:

	UINT_32 m_port;		    // port on the local machine

    public:

	    // constructors and destructor

	TcpipServerConnection 
	    (const BASE_SECURITY &security);
	
	virtual ~TcpipServerConnection
	    ();

	    // bind to a port and accept an incoming connection

	int setup
	    (UINT_32 port);

	SocketConnection *accept
	    (struct timeval stop);
};

    /* client-side Unix domain connection */

class UnixServerConnection: public SocketConnection
{
    protected:

	DynamicString m_socketFilename;		// Unix socket filename

    public:

	    // constructor and destructor

	UnixServerConnection 
	   (const BASE_SECURITY &security);

       ~UnixServerConnection 
	   ();

	    /* connect to the destination machine */

	int setup
	    (UINT_32 port);

	SocketConnection *accept
	    (struct timeval stop);
};

#endif /* _GEN_CONNECTION_H */
