/* Agent Tcl
   Bob Gray
   11 July 1995

   transmit.cc

   This file implmenets the routines that send and receive messages.

   Copyright (c) 1995, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef NO_PRAGMAS
#pragma implementation
#endif

#include "platPorting.h"
#include "agentId.h"
#include "genConnection.h"
#include "genEncrypt.h"
#include "genManager.h" 
#include "genSecurity.h"
#include "genTcpip.h"
#include "platTimers.h"
#include "genTransmit.h"
#include "genUtility.h"
#include "truefalse.h"

/* secureReceive

   Purpose: Receive an incoming message *performing any necessary decryption*

     Input: sockfd   = socket descriptor on which the message is arriving
		       (int)

	    set      = set of messages in which the message should appear
		       (class MESSAGE_SET &)

	    receiver = security characteristics of the receiver
		       (class BASE_SECURITY &)

    Output: The procedure returns one of the following codes.

	    MESSAGE_OK        = success
	    MESSAGE_SYS_ERROR = system error while receiving the message
	    MESSAGE_DECRYPT   = the message could not be decrypted
	    MESSAGE_INVALID   = the message was invalid

	    On success 

	    sender   = id of the encrypter and boolean flags that indicate
		       whether the message was encrypted and/or signed -- i.e,
		       the security characteristics of the sender
		       (class BASE_SECURITY &)

	    message  = pointer to a dynamically allocated MESSAGE structure
		       that contains the incoming message
		       (struct MESSAGE *&)
*/

int secureReceive (int sockfd, const MESSAGE_SET &set, const BASE_SECURITY &receiver, BASE_SECURITY &sender, MESSAGE *&message)
{
    int rc;
    MESSAGE *securedMessage = NULL;

	/* assume that we do not get a message */

    message = NULL;

	/* receive the secured message */

    if ((securedMessage = message_receive (sockfd, protectionMessages)) == NULL) {
	return (MESSAGE_INVALID);
    }

	/* unsecure the message */ 

    if ((rc = unsecure (set, receiver, *securedMessage, sender, message)) == MESSAGE_OK) {

	    /* and then receive the standard message if necessary */

	if (message == NULL) {

	    if ((message = message_receive (sockfd, set)) == NULL) {
		rc = MESSAGE_INVALID;
	    } else {
		rc = MESSAGE_OK;
	    }
	}
    }

    delete_check (securedMessage);
    return (rc);
}

/* message_receive

   Purpose: Receive a message from a file descriptor

     Input: sockfd = socket descriptor
            types  = possible message types

    Output: The procedure returns NULL on error.  Otherwise the procedure
            returns a pointer to a dynamically allocated MESSAGE structure 
            that contains the transmitted message.   
*/

MESSAGE *message_receive (int sockfd, const MESSAGE_SET &set)
{
    INT_32 netSize;
    INT_32 size;
    INT_32 fullSize; 
    char *buffer; 
    MESSAGE *message;

	/* get the size of the message */

    int rc;

    if ((rc = tcpip_read_long (sockfd, (UINT_32 *) &size)) < 0) {
	return NULL;
    } 

    if (size <= 0) {
	return NULL;
    }

    netSize  = htonl (size);
    fullSize = size + sizeof(UINT_32);

	/* allocate an empty MESSAGE */

    message = new MESSAGE (fullSize);
    buffer  = message -> getBuffer ();

	/* read the whole message */

    memcpy (buffer, (char *) &netSize, sizeof(UINT_32));

    if (tcpip_readn (sockfd, buffer + sizeof(UINT_32), size) < size) {
	delete (buffer);
	return NULL;
    }

	/* delinearize the message */

    if (message -> delinearize (set) < 0) {
	delete (message);
	message = NULL;
    }

	/* done */

    return (message);
}

MESSAGE *message_receive (Connection &connection, const MESSAGE_SET &set)
{
    int rc;
    UINT_32 netSize;
    INT_32 size;
    INT_32 fullSize; 
    char *buffer; 
    MESSAGE *message;

	/* get the size of the message */

    if ((rc = connection.read (sizeof(INT_32), (char *) &netSize)) < (int) sizeof(INT_32)) {
	return NULL;
    } 

    size = ntohl (netSize);

    if (size <= 0) {
	return NULL;
    }

    fullSize = size + sizeof(UINT_32);

	/* allocate an empty MESSAGE */

    message = new MESSAGE (fullSize);
    buffer  = message -> getBuffer ();

	/* read the whole message */

    memcpy (buffer, (char *) &netSize, sizeof(INT_32));

    if (connection.read (size, buffer + sizeof(UINT_32)) < size) {
	delete (buffer);
	return NULL;
    }

	/* delinearize the message */

    if (message -> delinearize (set) < 0) {
	delete (message);
	message = NULL;
    }

	/* done */

    return (message);
}

/* message_send

     Purpose: Write a message to a file descriptor

       Input: fd      = file descriptor
			(int)

              message = the message
			(class MESSAGE &)

      Output: The procedure returns -1 on error.  Otherwise the procedure
              writes the message to the file descriptor and returns 0.
*/

int message_send (int sockfd, MESSAGE &message)
{
    UINT_32 size;
    char *buffer;

	/* linearize the message */

    if (message.linearize () < 0) {
	return MESSAGE_SYS_ERROR;
    }

	/* write out the message */

    size    = message.getSize ();
    buffer  = message.getBuffer ();

    if (tcpip_writen (sockfd, buffer, size) < (int) size) {
	return MESSAGE_LOST_CONNECTION;
    }

    return MESSAGE_OK; 
}

int message_send (Connection &connection, MESSAGE &message)
{
    UINT_32 size;
    char *buffer;

	/* linearize the message */

    if (message.linearize () < 0) {
	return MESSAGE_SYS_ERROR;
    }

	/* write out the message */ 

    size    = message.getSize ();
    buffer  = message.getBuffer ();

    if (connection.write (size, buffer) < (int) size) {
	return MESSAGE_LOST_CONNECTION;
    }

    return MESSAGE_OK; 
}
